@@ -37,7 +37,6 @@
 #include <string.h>
 
 static void handle_get_items (DBusPendingCall *pending, void *user_data);
-gboolean _atspi_process_deferred_messages (gpointer data);
 
 static DBusConnection *bus = NULL;
 static GHashTable *live_refs = NULL;
@@ -733,7 +732,7 @@ typedef struct
   void *data;
 } BusDataClosure;
 
-static guint process_deferred_messages_id = -1;
+static GSource *process_deferred_messages_source = NULL;
 
 static void
 process_deferred_message (BusDataClosure *closure)
@@ -767,8 +766,8 @@ process_deferred_message (BusDataClosure *closure)
 
 static GQueue *deferred_messages = NULL;
 
-gboolean
-_atspi_process_deferred_messages (gpointer data)
+static gboolean
+process_deferred_messages (void)
 {
   static int in_process_deferred_messages = 0;
   BusDataClosure *closure;
@@ -783,13 +782,20 @@ _atspi_process_deferred_messages (gpointer data)
     dbus_connection_unref (closure->bus);
     g_free (closure);
   }
-  /* If data is NULL, assume that we were called from GLib */
-  if (!data)
-    process_deferred_messages_id = -1;
   in_process_deferred_messages = 0;
   return FALSE;
 }
 
+static gboolean
+process_deferred_messages_callback (gpointer data)
+{
+  if (process_deferred_messages ())
+    return G_SOURCE_CONTINUE;
+
+  process_deferred_messages_source = NULL;
+  return G_SOURCE_REMOVE;
+}
+
 static DBusHandlerResult
 defer_message (DBusConnection *connection, DBusMessage *message, void *user_data)
 {
@@ -801,12 +807,12 @@ defer_message (DBusConnection *connection, DBusMessage *message, void *user_data
 
   g_queue_push_tail (deferred_messages, closure);
 
-  if (process_deferred_messages_id == -1)
+  if (process_deferred_messages_source == NULL)
   {
-    GSource *source = g_idle_source_new ();
-    g_source_set_callback (source, _atspi_process_deferred_messages, NULL, NULL);
-    process_deferred_messages_id = g_source_attach (source, atspi_main_context);
-    g_source_unref (source);
+    process_deferred_messages_source = g_idle_source_new ();
+    g_source_set_callback (process_deferred_messages_source,
+                           process_deferred_messages_callback, NULL, NULL);
+    g_source_attach (process_deferred_messages_source, atspi_main_context);
   }
 
   return DBUS_HANDLER_RESULT_HANDLED;
@@ -1131,7 +1137,7 @@ _atspi_dbus_call (gpointer obj, const char *interface, const char *method, GErro
                                            type, args);
   va_end (args);
   check_for_hang (NULL, &err, aobj->app->bus, aobj->app->bus_name);
-  _atspi_process_deferred_messages ((gpointer)TRUE);
+  process_deferred_messages ();
   if (dbus_error_is_set (&err))
   {
     g_set_error(error, ATSPI_ERROR, ATSPI_ERROR_IPC, "%s", err.message);
@@ -1188,7 +1194,7 @@ out:
   va_end (args);
   if (msg)
     dbus_message_unref (msg);
-  _atspi_process_deferred_messages ((gpointer)TRUE);
+  process_deferred_messages ();
   if (dbus_error_is_set (&err))
   {
     /* TODO: Set gerror */
@@ -1239,7 +1245,7 @@ _atspi_dbus_get_property (gpointer obj, const char *interface, const char *name,
   reply = dbind_send_and_allow_reentry (aobj->app->bus, message, &err);
   check_for_hang (reply, &err, aobj->app->bus, aobj->app->bus_name);
   dbus_message_unref (message);
-  _atspi_process_deferred_messages ((gpointer)TRUE);
+  process_deferred_messages ();
   if (!reply)
   {
     // TODO: throw exception
@@ -1302,7 +1308,7 @@ _atspi_dbus_send_with_reply_and_block (DBusMessage *message, GError **error)
   dbus_error_init (&err);
   set_timeout (app);
   reply = dbind_send_and_allow_reentry (bus, message, &err);
-  _atspi_process_deferred_messages ((gpointer)TRUE);
+  process_deferred_messages ();
   dbus_message_unref (message);
   if (dbus_error_is_set (&err))
   {
@@ -1650,6 +1656,16 @@ atspi_set_timeout (gint val, gint startup_time)
 void
 atspi_set_main_context (GMainContext *cnx)
 {
+  if (atspi_main_context == cnx)
+    return;
+  if (process_deferred_messages_source != NULL)
+  {
+    g_source_destroy (process_deferred_messages_source);
+    process_deferred_messages_source = g_idle_source_new ();
+    g_source_set_callback (process_deferred_messages_source,
+                           process_deferred_messages_callback, NULL, NULL);
+    g_source_attach (process_deferred_messages_source, cnx);
+  }
   atspi_main_context = cnx;
   atspi_dbus_connection_setup_with_g_main (atspi_get_a11y_bus (), cnx);
 }
